/*************************************************************************
 *
 *
 *    (c) Copyright Olimex 2011
 *
 *    File name   : MAG3310.c
 *    Description : API for the Freescale MAG3110 3-Axis, Digital Magnetometer
 *
 *    History :
 *    1. Date        : 14 July 2011
 *       Author      : Aleksandar Mitev
 *       Description : Create
 *
 **************************************************************************/
#include "Compiler.h"
#include "GenericTypeDefs.h"
#include "MAG3110.h"

/* DEFINE LOCAL TYPES HERE */

/* DEFINE LOCAL CONSTANTS HERE */
#define MAG3110_SDA_TRIS TRISBbits.TRISB5
#define MAG3110_SDA_LAT LATBbits.LATB5
#define MAG3110_SDA_PORT PORTBbits.RB5

#define MAG3110_SCL_TRIS TRISBbits.TRISB4
#define MAG3110_SCL_LAT LATBbits.LATB4
#define MAG3110_SCL_PORT PORTBbits.RB4

#define MAG3110_INT_TRIS TRISCbits.TRISC7
#define MAG3110_INT_PORT PORTCbits.RC7

#define MAG3110_SLAVE_ADDR 0b00011100

#define ACK 0
#define NACK 1

/* DECLARE EXTERNAL VARIABLES HERE */

/* DEFINE LOCAL MACROS HERE */

/* DEFINE LOCAL VARIABLES HERE */
static WORD_VAL CompassX; // measured compass raw value
static WORD_VAL CompassY; // measured compass raw value
static WORD_VAL CompassZ; // measured compass raw value
static BYTE Temperature; // measured tempr raw value

/* DECLARE EXTERNAL VARIABLES HERE */

/* DECLARE LOCAL FUNCTIONS HERE */
static void local_I2C_Delay(void);
static void local_I2C_Start(void);
static void local_I2C_Stop(void);
static BYTE local_I2C_ReadByte(char ack);
static char local_I2C_WriteByte(BYTE data); // returns ack state

/* DEFINE FUNCTIONS HERE */

/******************************************************************************
* Description: MAG3110_Initialize(..) - initializes pins and registers of the MAG3310
*				Also enables automatic sampling
* Input: 	none
* Output: 	none
* Return:	0 if sucessfully initialized, -1 if error occured 
*******************************************************************************/
char MAG3110_Initialize(void)
{
	char result = 0;
	BYTE id;
	
	// the idea is to toggle pins as inputs/outputs to emulate I2C open drain mode of operation!	
	MAG3110_SDA_TRIS = 1;
	MAG3110_SCL_TRIS = 1;
	
	MAG3110_SDA_LAT = 0;
	MAG3110_SCL_LAT = 0;
	
	MAG3110_INT_TRIS = 1;
		
	local_I2C_Delay();
	if( !(MAG3110_SCL_PORT && MAG3110_SDA_PORT) )
		return -1;

	do {
		// check the ID for match
		local_I2C_Start();
		result |= local_I2C_WriteByte(MAG3110_SLAVE_ADDR | 0x00);
		result |= local_I2C_WriteByte(0x07);
		local_I2C_Stop();
		
		// is there such a chip on the line at all?
		if(result) break;
				
		local_I2C_Start();
		result |= local_I2C_WriteByte(MAG3110_SLAVE_ADDR | 0x01);
		id = local_I2C_ReadByte(NACK);
		local_I2C_Stop();
		
		// bus error or ID mismatch?
		if(result || !(id == 0xC4))
			break;

		// configure device here
		local_I2C_Start();
		result |= local_I2C_WriteByte(MAG3110_SLAVE_ADDR | 0x00);
		result |= local_I2C_WriteByte(0x10);
		result |= local_I2C_WriteByte(0b00111000); // CTRL_REG1
		result |= local_I2C_WriteByte(0b10100000); // CTRL_REG2
		local_I2C_Stop();
		
		// is there an error configuring device
		if(result) break;
		
		// enable device
		local_I2C_Start();
		result |= local_I2C_WriteByte(MAG3110_SLAVE_ADDR | 0x00);
		result |= local_I2C_WriteByte(0x10);
		result |= local_I2C_WriteByte(0b00111001); // CTRL_REG1
		local_I2C_Stop();
		
		// is there an error enabling device
		if(result) break;

	} while(0);
	
	CompassX.Val = 0;
	CompassY.Val = 0;
	CompassZ.Val = 0;
	Temperature = 0;

	return (result ? -1 : 0);
}

/******************************************************************************
* Description: MAG3110_ReadAll(..) - reads the results of the last sampling
* Input: 	none
* Output: 	none
* Return:	-1 on error, 0 if successful
*******************************************************************************/
char MAG3110_ReadAll(MAG3110_VALUES *pVal)
{
	BYTE dummy;
	char newFlag = 0;
	char result = 0;
	
	if(MAG3110_INT_PORT == 1) { //  new values available
		
		// read the axes
		local_I2C_Start();
		result |= local_I2C_WriteByte(MAG3110_SLAVE_ADDR | 0x00);
		result |= local_I2C_WriteByte(0x01);
		local_I2C_Stop();
		local_I2C_Start();
		result |= local_I2C_WriteByte(MAG3110_SLAVE_ADDR | 0x01);
		CompassX.byte.HB = local_I2C_ReadByte(ACK);
		CompassX.byte.LB = local_I2C_ReadByte(ACK);
		CompassY.byte.HB = local_I2C_ReadByte(ACK);
		CompassY.byte.LB = local_I2C_ReadByte(ACK);
		CompassZ.byte.HB = local_I2C_ReadByte(ACK);
		CompassZ.byte.LB = local_I2C_ReadByte(ACK);
		CompassZ.byte.LB = local_I2C_ReadByte(NACK);
		local_I2C_Stop();

		// read the temperature
		local_I2C_Start();
		result |= local_I2C_WriteByte(MAG3110_SLAVE_ADDR | 0x00);
		result |= local_I2C_WriteByte(0x0F);
		local_I2C_Stop();
		local_I2C_Start();
		result |= local_I2C_WriteByte(MAG3110_SLAVE_ADDR | 0x01);
		Temperature = local_I2C_ReadByte(NACK);
		local_I2C_Stop();

		// was there an error getting the values
		if(result)
			return -1;
			
		newFlag = 1;
	}
	
	// no idea how to process values, just copy over to the return struct
	pVal->x = CompassX.Val;
	pVal->y = CompassY.Val;
	pVal->z = CompassZ.Val;
	pVal->tempr = Temperature;	
		
	return newFlag;
}

/* local functions */

static void local_I2C_Delay(void)
{
	UINT d = 1000;
	while(d--) {
		Nop();
	}	
}	

static void local_I2C_Start(void)
{
	MAG3110_SDA_TRIS = 0;
	local_I2C_Delay();
	MAG3110_SCL_TRIS = 0;
	local_I2C_Delay();
	
}

static void local_I2C_Stop(void)
{
	MAG3110_SDA_TRIS = 0;
	local_I2C_Delay();
	MAG3110_SCL_TRIS = 1;
	local_I2C_Delay();
	MAG3110_SDA_TRIS = 1;
	local_I2C_Delay();
}

static BYTE local_I2C_ReadByte(char ack)
{
	BYTE data = 0;
	char i;
	
	MAG3110_SDA_TRIS = 1;
	for(i = 0; i < 8; i++) {
		local_I2C_Delay();
		MAG3110_SCL_TRIS = 1;
		local_I2C_Delay();
		data |= MAG3110_SDA_PORT & 0x01;
		if(i != 7)
			data <<= 1;
		MAG3110_SCL_TRIS = 0;
	}
	
	// read the ack
	local_I2C_Delay();
	MAG3110_SDA_TRIS = ack;
	local_I2C_Delay();
	MAG3110_SCL_TRIS = 1;
	local_I2C_Delay();
	MAG3110_SCL_TRIS = 0;
	local_I2C_Delay();
	
	return data;
}

// returns ack state, 0 means acknowledged
static char local_I2C_WriteByte(BYTE data)
{
	char i;

	// send the 8 bits
	for(i = 0; i < 8; i++) {
		MAG3110_SDA_TRIS = (data & 0x80) ? 1 : 0;
		data <<= 1;
		local_I2C_Delay();
		MAG3110_SCL_TRIS = 1;
		local_I2C_Delay();
		MAG3110_SCL_TRIS = 0;
	}
	
	// read the ack
	MAG3110_SDA_TRIS = 1;
	local_I2C_Delay();
	MAG3110_SCL_TRIS = 1;
	local_I2C_Delay();
	i = MAG3110_SDA_PORT;
	MAG3110_SCL_TRIS = 0;
	local_I2C_Delay();
	
	return i;
}


